/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   BSD - specific code
 *   this file is a part of xtrace project
 *
 * $Id: bsd.c,v 1.3 2002/04/23 07:56:40 drozd Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ptrace.h>
#include <sys/wait.h>

#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#include "xtrace.h"
#include "bsd.h"
#include "elf.h"

static struct bsddata __data;		/* bsd - specific data */
static void fini(struct opt *);

int
xtr_stop_app(struct opt *op)
{
	struct bsddata *bdp = (struct bsddata *) op->data;

	if (ioctl(bdp->procfd, PIOCWAIT, &bdp->status) < 0) {
		perror("ioctl(PIOCWAIT)"); return (-1);
	}

	if (bdp->status.why == S_EXIT) {
		op->flags |= XTR_CHILDEXITED;
		fini((struct opt *)op);
		exit(0);
	}

	return (0);
}

int
xtr_resume_app(struct opt *op)
{
	int val = 0;
	struct bsddata *bdp = (struct bsddata *) op->data;
	
	if (ioctl(bdp->procfd, PIOCCONT, val) < 0) {
		perror("ioctl(PIOCCONT)"); return (-1);
	}

	return (0);
}

/*
 * Initialize tracer
 * Parameters:
 *    op - current options
 *    slist - comma separated list of syscalls to trace
 *    flist - comma separated list of functions to trace
 * Return:
 *    0 on success, < 0 on error
 */
int
xtr_init(struct opt *op, char *slist, char *flist)
{
	char fname[128];
	struct bsddata *bdp = &__data;
	struct procfs_status status;
	unsigned long flags;

	UNUSED(slist);
	UNUSED(flist);
	
	/* register cleanup and SIGCLD handler functions, setup data ptr */
	op->fini = fini;
	op->data = &__data;
	op->child = NULL;

	/* open a control file */
	(void) snprintf(fname, sizeof(fname), "/proc/%d/mem", op->pid);
	if ((bdp->procfd = open(fname, O_RDWR)) < 0) {
		perror("xtr_init: open(/proc/pid/mem, O_RDWR"); exit(1);
	}

	/* open registers file */
	(void) snprintf(fname, sizeof(fname), "/proc/%d/regs", op->pid);
	if ((bdp->regsfd = open(fname, O_RDONLY)) < 0) {
		perror("xtr_init: open(/proc/pid/regs, O_RDWR"); exit(1);
	}

	/* remember old flags */
	if (ioctl(bdp->procfd, PIOCSTATUS, &status) == -1) {
		perror("xtr_init: ioctl(PIOCSTATUS)"); exit(1);
	}
	bdp->oevnts = status.events;

	/* continue the process on last close of /proc/pid/mem */
	flags = PF_LINGER;
	if (ioctl(bdp->procfd, PIOCSFL, flags) == -1) {
		perror("xtr_init: ioctl(PF_LINGER)"); exit(1);
	}
	
	/* stop on those events  */
	bdp->evnts = S_EXEC | S_SCE | S_SCX | S_CORE | S_EXIT;
	if (ioctl(bdp->procfd, PIOCBIS, bdp->evnts) == -1) {
		perror("xtr_init: ioctl(PIOCBIS)"); exit(1);
	}
	
	return (0);
}

/*
 * Dump memory content from process's address space
 */
void
xtr_read_mem(const struct opt *op, vaddr_t va, char *buf, int len)
{
	struct bsddata *dp = (struct bsddata *) op->data;

	(void) lseek(dp->procfd, va, SEEK_SET);
	
	if (read(dp->procfd, buf, len) != len)
		perror("xtr_read_mem: read(va)");
}


/*
 * Get load address for given ELF object
 * Parameters:
 *    op - ptr to options
 *    fname - file name of the ELF object
 * Return:
 *    base load address
 */
vaddr_t
xtr_get_load_addr(const struct opt *op, const char *fname)
{
	UNUSED(op);
	UNUSED(fname);

	return (0U);
}

void
xtr_before_exec(struct opt *op)
{
	int fd;
	long evnts;
	char fname[128];

	UNUSED(op);

	(void) snprintf(fname, sizeof(fname), "/proc/%d/mem", getpid());

	if (access(fname, R_OK | W_OK) < 0) {
		perror("xtr_before_exec: access(/proc/<pid>/mem)"); exit(1);
	}

	/* open a control file */
	if ((fd = open(fname, O_RDWR)) < 0) {
		perror("xtr_brefore_exec: "
		"open(/proc/<pid>/mem, O_RDWR | O_EXCL)"); exit(1);
	}

	/* stop on those events  */
	evnts = S_EXEC;
	if (ioctl(fd, PIOCBIS, evnts) == -1) {
		perror("xtr_before_exec: ioctl(PIOCBIS)"); exit(1);
	}

	(void) close(fd);
}

/*
 * Cleanup function. Must handle Linux-specific stuff.
 * See what xtr_fini() does before patching this code
 */
void
fini(struct opt *op)
{
	struct bsddata *bdp = (struct bsddata *) op->data;
		
	if (!(op->flags & XTR_CHILDEXITED)) {
		if (ioctl(bdp->procfd, PIOCBIS, bdp->oevnts) == -1)
			perror("restoring event mask: ioctl(PIOCBIS)");
	}

	(void) close(bdp->procfd);
	(void) close(bdp->regsfd);
}
