/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   dump call routines
 *   this file is a part of xtrace project
 *
 * $Id: dump.c,v 1.1 2002/04/23 07:56:40 drozd Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/syscall.h>

#include <netinet/in.h>

#include <stdio.h>

#include "xtrace.h"

static void dumparg(struct opt *op, enum argtype type, vaddr_t va);
static void backtrace(struct opt *, struct calldata *);

void
dumparg(struct opt *op, enum argtype type, vaddr_t va)
{
	const char *fmt = NULL;
	struct stat st;
	struct sockaddr sa;
	struct sockaddr_in *sin;
	unsigned char *ucp;
	char buf[256];

	switch (type) {
		case STAT:
			fmt = "%s";
			(void) xtr_read_mem(op, va, (void *)&st, sizeof(st));
			(void) snprintf(buf, sizeof(buf),
			"dev: %d ino: %d mod: %d nlink: %d uid: %d gid: %d"
			" atime: %lu ctime: %lu mtime: %lu size: %ld",
			st.st_dev, st.st_ino, st.st_mode, st.st_nlink,
			st.st_uid, st.st_gid, st.st_atime,
			st.st_ctime, st.st_mtime, st.st_size);
			buf[sizeof(buf) - 1] = '\0';
			va = (vaddr_t) buf;
			break;
		case SOCKADDR:
			fmt = "%s";
			(void) xtr_read_mem(op, va, (void *)&sa, sizeof(sa));
			if (sa.sa_family == PF_INET) {
				sin = (struct sockaddr_in *) &sa;
				ucp = (unsigned char *) &sin->sin_addr;
				(void) snprintf(buf, sizeof(buf),
				"{family: PF_INET port: %d addr: %d.%d.%d.%d}",
				ntohs(sin->sin_port), *ucp, *(ucp + 1),
				*(ucp + 2), *(ucp + 3));
			} else {
				(void) snprintf(buf, sizeof(buf),
				"{ family: %d, data: %s }",
				sa.sa_family, sa.sa_data);
			}
			buf[sizeof(buf) - 1] = '\0';
			va = (vaddr_t) buf;
			break;
		/* FIXME */
		case MSGHDR:
		case SIGALTSTACK:
		case IOCTL:
		case ARGV:
		case ITIMER:
		case FCNTL:
		case TIMEVAL:
		case TIMEZONE:
		case IOVEC:
		case STATFS:
		case RTPRIO:
		case RLIMIT:
		case TIMESPEC:
		case MODSTAT:
		case KLDSTAT:
		case SHEDPARAM:
		case SIGSET:
		case SIGACTION:
		case KEVENT:
		case HEX: fmt = "%#x"; break;
		case INT: fmt = "%d"; break;
		case STRING:
			fmt = "\"%s\"";
			(void) xtr_read_mem(op, va, buf, sizeof(buf));
			buf[sizeof(buf) - 1] = '\0';
			va = (vaddr_t) buf;
			break;
		default: break;
	}

	op->ncols += printf(fmt, va);
}

/*
 * Dump information about current call
 * Parameters:
 *   sp - ptr to struct prstatus, should be read from /proc/<pid>/status
 * Return:
 *   void
 */
void
xtr_dump_call(struct opt *op)
{
	const struct syscall *scp;
	struct calldata cd;
	char **fnames;
	int i;

	(void) memset(&cd, 0, sizeof(struct calldata));
	if (xtr_get_call_data(op, &cd) != 0)
		return;

	if (op->flags & XTR_INSYSEXIT) {

		/* print syscall's return value */
		(void) printf(" = %u ", cd.retval);
		if (op->flags & XTR_TIMESTAMP)
			xtr_dump_time();
		(void) putchar('\n');

			
		/* rebuild symbol table after execve() */
		if (op->scno == SYS_execve) {
			xtr_symtab_destroy();
			fnames = xtr_get_elfs(op->pid);
			xtr_symtab_build(op, fnames);
		}
	
		/* dump backtrace if requested */
		if (op->flags & XTR_BACKTRACE)
			backtrace(op, 0);

	} else if (op->flags & XTR_INSYSENTER) {
		
		op->ncols = 0;
		
		scp = xtr_lookup_syscall(op->scno);
		op->ncols += printf("%s(", scp->name);
		
		/* print out the syscall's arguments */
		for (i = 0; i < scp->nargs; i++) {
			dumparg(op, scp->args[i], *cd.argv[i]);
			if (i < scp->nargs -1)
				op->ncols += putchar(',');
		}

		op->ncols += putchar(')');
		if (op->scno == 1)
			putchar('\n');
	}
}

/*
 * Dump call stack
 * Parameters:
 *    op - ptr to options
 * Return:
 *    void
 */
void
backtrace(struct opt *op, struct calldata *cdp)
{
	ptrdiff_t diff;		/* offset from the function prologue */
	const struct symbol *symp;
	struct calldata cd;
	int leaf = 0;
	vaddr_t ip;

	/* set the frame pointer to 0 */
	if (cdp == 0) {
		cd.fp = (vaddr_t) 0;
		cdp = &cd;
		++leaf;
	}

	if (xtr_get_call_data(op, cdp) != 0)
		return;

	if((cdp->rta == 0U) || (cdp->ofp == 0U))
		return;

	ip = leaf == 0 ? cdp->rta : cdp->ip;
	symp = xtr_find_symbol(ip, &diff);
	if (!symp)
		return;

	/* dump call data */
	(void) printf("\t%-8x %s (", ip, symp->name);
	for (; *cdp->argv != 0; cdp->argv++) {
		(void) printf("%x", **cdp->argv);
		if (op->flags & XTR_DUMPMEM) {
			(void) xtr_read_mem(op, **cdp->argv, NULL, 0x10);
			/* FIXME: dump content onto stdout */
		}
		(void) putchar(',');
	}
	(void) printf(") + %x", diff);

	(void) putchar('\n');
	
	/* do that recursively  - shift one frame up */
	cdp->fp = cdp->ofp;
	backtrace(op, cdp);
}


