/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   Linux - specific code
 *   this file is a part of xtrace project
 *
 * $Id: linux.c,v 1.2 2002/04/21 17:21:10 drozd Exp $
 * $Log: linux.c,v $
 * Revision 1.2  2002/04/21 17:21:10  drozd
 * scall, fcall removed from struct opt
 *
 * Revision 1.1  2002/02/24 18:25:29  drozd
 * change in interface - svr4 stuff is broken now
 *
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ptrace.h>
#include <sys/wait.h>

#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>

#include "xtrace.h"
#include "linux.h"
#include "elf.h"

#include "linux-syscalls.c"

static struct linuxdata __data;		/* linux - specific data */

static void fini(struct opt *);
static void childhandler(int);

#define CATCHSYSCALLS(op)						\
do {									\
	if (op->flags & XTR_TRACESYS) {					\
		if (ptrace(PTRACE_SYSCALL, op->pid, 1, 0) != 0) {	\
			perror("ptrace(PTRACE_SYSCALL)"); return -1;	\
		}							\
	}								\
} while (0)

/*
 * Initialize tracer
 * Parameters:
 *    op - current options
 *    slist - comma separated list of syscalls to trace
 *    flist - comma separated list of functions to trace
 * Return:
 *    0 on success, < 0 on error
 */
int
xtr_init(struct opt *op, char *slist, char *flist)
{
	/* register cleanup and SIGCLD handler functions, setup data ptr */
	op->fini = fini;
	op->child = childhandler;
	op->data = &__data;
	op->scall = dump_linux_syscall;

	(void) signal(SIGCLD, childhandler);

	/* attach to a process */
	if (ptrace(PTRACE_ATTACH, op->pid, 0, 0) != 0) {
		perror("ptrace(PTRACE_ATTACH)"); return -1;
	}

	return 0;
}

/*
 * Start a kid
 */
pid_t
xtr_start_app(char *const *argv, char *const *envp)
{
	return -1;
}

/* get list of dependant ELF binaries */
char *const *
xtr_get_elfs(pid_t pid)
{
	Elf32_Ehdr *ep;
	int fd, nbytes, i;
	char link[512], fname[512];
	static char *fnames[128];
	char *const *depends;

	/* read 'exe' symlink */
	(void) snprintf(link, sizeof(link), "/proc/%d/exe", pid);
	nbytes = readlink(link, fname, sizeof(fname));
	fname[nbytes] = '\0';

	if ((fd = open(fname, O_RDONLY)) < 0) {
		perror("xtr_get_elfs open()");
		return 0;
	}
	
	if ((ep = elf_init(fd)) == 0)
		return 0;

	fnames[0] = strdup(fname);
	fnames[1] = 0;

	depends = elf_get_depends(ep);
	for (i = 0; depends != 0 && depends[i] != 0; i++)
		fnames[i + 1] = depends[i];

	/* null - terminate the array */
	fnames[i + 1] = NULL;

	
	elf_fini(ep);

	return fnames;
}

/*
 * Stop on event of interest
 */
int
xtr_stop_app(const struct opt *op)
{
	int status;

	(void) wait(&status);

	if (WIFEXITED(status))
		exit(0);

	return 0;
}

/*
 * Resume process
 */
int
xtr_resume_app(const struct opt *op)
{
	/* catch syscalls if requested */
	CATCHSYSCALLS(op);

	return 0;
}

/*
 * Dump memory content from process's address space
 */
void
xtr_dump_mem(const struct opt *op, vaddr_t va, int len)
{
}

/*
 * Cleanup function. Must handle Linux-specific stuff.
 * See what xtr_fini() does before patching this code
 */
void
fini(struct opt *op)
{
	int status;

	if (waitpid(op->pid, &status, WNOHANG) > 0) {
		/* detach from a process blindly */
		if (ptrace(PTRACE_DETACH, op->pid, 0, 0) != 0) {
			perror("ptrace(PTRACE_DETACH)");
		}
	}
	(void) putchar('\n');
}

/*
 * SIGCLD handler
 */
void
childhandler(int signo)
{
	//pid_t pid;
	//int status;

	//pid = wait(&status);
	//printf("in child handler, status %d\n", status);
	//ptrace(PTRACE_SYSCALL, pid, 0, 0);
}


/*
 * prepare child to be traced by xtrace
 */
void
xtr_before_exec(struct opt *op)
{
	UNUSED(op);
}

/*
 * Get load address for given ELF object
 * Parameters:
 *    op - ptr to options
 *    fname - file name of the ELF object
 * Return:
 *    base load address
 */
vaddr_t
xtr_get_load_addr(const struct opt *op, const char *fname)
{
	char rbuf[BUFSIZ], map[128];
	FILE *fp;
	char *mapname, *nl;
	vaddr_t retval;
	int found;

	/* try to open map filename */
	(void) snprintf(map, sizeof(map), "cat /proc/%d/maps", op->pid);

	if ((fp = popen(map, "r")) == 0) {
		perror("xtr_get_load_addr fopen()");
		return 0;
	}

	/* find mapping */
	for (retval = 0, found = 0; fgets(rbuf, sizeof(rbuf), fp) != NULL; ) {
		if ((mapname = strchr(rbuf, '/')) == 0)
			continue;

		/* change newline to '\0' */
		if ((nl = strchr(mapname, '\n')) != 0)
			*nl = '\0';

		if (strcmp(mapname, fname) != 0)
			continue;

		/* found. grab load address */
		rbuf[8] = 0;
		(void) sscanf(rbuf, "%x", &retval);
		break;
	 }

	(void) pclose(fp);

	return retval;
}
