/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   Main module for xtrace utility
 *   this file is a part of xtrace project
 *
 * $Id: xtrace.c,v 1.7 2002/04/23 07:56:40 drozd Exp $
 */

#include <sys/types.h>
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include "xtrace.h"

static const char __version[] =
"$Id: xtrace.c,v 1.7 2002/04/23 07:56:40 drozd Exp $";

/* options */
static struct opt __opt;

static void prepare(struct opt *, char *, char *);
static void start_app(struct opt *, char *const *, char *const *);
static void fini(int);

/*
 * Print usage information and exit
 * Parameters:
 *   pname - executable's name
 * Return value:
 *   void
 */
void usage(const char *pname)
{
	(void) fprintf(stderr, "usage:\n%s [OPTIONS] [filename ...]\n"
	    "OPTIONS:\n"
	    "-p <pid>\tPID of a process to be traced\n"
	    "-h\t\tthis help\n"
	    "-V\t\tprint version and exit\n"
	    "-S\t\tdump stack trace\n"
	    "-T\t\tdump per-thread info\n"
	    "-o <file>\tredirect output to file\n"
	    "-t\t\tput timestamp for every syscall\n"
	    "-v\t\tshow memory content for valid addresses\n"
	    "-s <syscall[,syscall2...]>\ttrace specific syscalls\n"
	    "-f <func1,[func2...]|all>\ttrace functions\n"
	    "-L\t\tdump list of functions from executable and exit\n"
	    "Note: -v, -S, -f options affect perfomance seriously!\n",
	    pname);
	exit(1);
}

/*
 * jump point for crt1.o code
 */
int
main(int argc, char *const *argv, char *const *envp)
{
	int ch, retval;
	extern char *optarg;
	extern int optind;
	char *syslist = 0, *funclist = 0;

	while ((ch = getopt(argc, argv, "vVtTLSs:f:p:o:")) != -1)
		switch (ch) {
			case 'p':
				__opt.pid = (pid_t) atoi(optarg);
				break;
			case 't':
				__opt.flags |= XTR_TIMESTAMP;
				break;
			case 's':
				__opt.flags |= XTR_TRACESYS;
				syslist = optarg;
				break;
			case 'o':
				(void) freopen(optarg, "w", stdout);
				break;
			case 'v':
				__opt.flags |= XTR_DUMPMEM;
				break;
			case 'f':
				__opt.flags |= XTR_TRACEFUNC ;
				funclist = optarg;
				break;
			case 'S':
				__opt.flags |= XTR_BACKTRACE;
				break;
			case 'T':
				__opt.flags |= XTR_DUMPTHREADS;
				break;
			case 'L':
				__opt.flags |= XTR_FUNCLIST;
				break;
			case 'V':
				(void) printf("%s\n", __version);
				_exit(0);
				break;
			case 'h':
				/* FALLTHROUGH */
			default:
				usage(*argv);
				/* NOTREACHED */
				break;
		}

	/* check child's PID sanity and start the child if requested */
	if (argc != optind) {
		start_app(&__opt, &argv[optind], envp);
	} else if (__opt.pid <= (pid_t) 0) {
		(void) fprintf(stderr, "invalid pid %d\n", (int)__opt.pid);
		exit(1);
	}

	prepare(&__opt, syslist, funclist);
	
	/* infinite trace loop */
	for (retval = -1; retval != 0; ) {

		/* stop on event of interest */
		(void) xtr_stop_app(&__opt);

		/* show it */
		xtr_dump_call(&__opt);

		/* resume process */
		(void) xtr_resume_app(&__opt);
	}

	return 0;
}

/*
 * Signal handler, and atexit handler in one
 * Parameters:
 *   signo - signal number caught
 * Return:
 *   void
 */
void
fini(int signo)
{
	/* call platform specific cleanup */
	if (__opt.fini != 0)
		(*__opt.fini)(&__opt);

	xtr_symtab_destroy();

	(void) fflush(stderr);
	(void) fflush(stdout);

	_exit(0);
}

/*
 * Print out a time
 * Parameters, Return: void
 */
void
xtr_dump_time(void)
{
	struct timeval tv = { 0U, 0U };

	(void) gettimeofday(&tv, 0);
	(void) printf("(%lu.%lu)", tv.tv_sec, tv.tv_usec);
}

/*
 * Platform-independet xtr_init()
 * Parameters:
 *    op - ptr to options
 *    slist - comma separated list of syscalls to trace
 *    flist - comma separated list of functions to trace
 * Return: void
 */
void
prepare(struct opt *op, char *slist, char *flist)
{
	 char **fnames;		/* ELF files to be processed */

	/* trace syscalls by default */
	if (!(op->flags & XTR_TRACESYS) && !(op->flags & XTR_TRACEFUNC)) {
		op->flags |= XTR_TRACESYS;
		slist = "all";
	}

	(void) atexit((void (*)(void)) fini);

	(void) signal(SIGINT, fini);
	(void) signal(SIGTERM, fini);
	(void) signal(SIGQUIT, fini);

	/* call platform dependent init function */
	if (xtr_init(op, slist, flist) < 0) {
		(void) fprintf(stderr, "error init xtrace\n");
		exit(1);
	}

	/* set SIGCLD handler to fini if xtr_init() didn't set it */
	if (op->child != 0) {
		(void) signal(SIGCHLD, op->child);
	}

	/*
	 * FIXME: I must find a way to signal to parent when child has
	 * execve() completed so to rebuild symbol table
	 */

	/* fill up symbol table */
	if ((op->flags & XTR_TRACEFUNC) || (op->flags & XTR_BACKTRACE)) {
		fnames = xtr_get_elfs(op->pid);
		(void) xtr_symtab_build(op, fnames);
	}

	/* dump function list and exit if requested */
	if (op->flags & XTR_FUNCLIST) {
		struct symbol **symp;
		
		if (op->flags & XTR_STARTCHLD) {
			fnames[0] = op->exename;
			fnames[1] = NULL;
			xtr_symtab_build(op, fnames);
		}
		
		for (symp = 0; (symp = xtr_get_next_sym(symp)) != 0; )
			if ((*symp)->flags & SYM_REFERENCED)
				(void) printf("%-8x %s\n", (*symp)->va,
				     (*symp)->name);
		exit(0);
	}
}

/* 
 * Start the application, ordering to stop after execve()
 * Parameters:
 *    argv - arguments vector
 *    envp - environment
 * Return: void
 */
void
start_app(struct opt *op, char *const *argv, char *const *envp)
{
	op->flags |= XTR_STARTCHLD;
	op->exename = argv[0];
	op->pid = fork();

	/* if we're a child, set correct PID  */
	if (op->pid == (pid_t) 0)
		op->pid = getpid();
	else
		return;

	/* set up the exec handler */
	xtr_before_exec(op);

	(void) execve(argv[0], argv, envp);

	/* if we're here, execve() failed */
}
