/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   API definition for xtrace utility, see README for details.
 *   this file is a part of xtrace project
 *
 * List of platform-specific functions, that have to be defined for every
 * platform specific module:
 *
 *   o xtr_init
 *   o xtr_start_app
 *   o xtr_stop_app
 *   o xtr_resume_app
 *   o xtr_get_call_data
 *   o xtr_dump_mem
 *   o xtr_get_load_addr
 *   o xtr_get_sp
 *
 * $Id: xtrace.h,v 1.6 2002/04/23 07:56:40 drozd Exp $
 * $Log: xtrace.h,v $
 * Revision 1.6  2002/04/23 07:56:40  drozd
 * backtrace support. svr4 & linux are broken
 *
 * Revision 1.5  2002/04/22 11:25:04  drozd
 *
 * os-syscalls deprecated now. all syscalls in syscall.c
 *
 * Revision 1.4  2002/04/21 17:21:10  drozd
 * scall, fcall removed from struct opt
 *
 * Revision 1.3  2002/02/24 18:25:29  drozd
 * change in interface - svr4 stuff is broken now
 *
 * Revision 1.2  2002/02/21 13:49:56  drozd
 * function trace enabled
 *
 * Revision 1.1.1.1  2002/02/21 08:55:10  drozd
 * import
 *
 */

#ifndef __XTRACE__H
#define __XTRACE__H

#include <stdlib.h>
#include <unistd.h>
#include <stddef.h>

typedef unsigned vaddr_t;

#define UNUSED(arg)	(arg) = (arg)

/* options */
struct opt {
	pid_t pid;		/* process's PID */
	unsigned flags;		/* see below for valid values */
	int scno;		/* syscall number */
	int ncols;		/* characters printed */
	char *exename;		/* executable's full path name */

	/* platform - specific stuff */
	void *data;			/* generic data */
	void (*fini)(struct opt *);	/* cleanup */
	void (*child)(int);		/* SIGCLD handler ??? */
};

#define XTR_BACKTRACE	(1 << 0)	/* dump per-thread backtrace */
#define XTR_TIMESTAMP	(1 << 1)	/* dump timestamp */
#define XTR_FUNCLIST	(1 << 2)	/* dump function list and exit */
#define XTR_DUMPMEM	(1 << 3)	/* dump memory content for parameters */
#define XTR_TRACESYS	(1 << 4)	/* trace system calls */
#define XTR_TRACEFUNC	(1 << 5)	/* trace functions */
#define XTR_DUMPTHREADS	(1 << 6)	/* dump per-thread info */
#define XTR_INSYSENTER	(1 << 7)	/* stopped in syscall entry */
#define XTR_INSYSEXIT	(1 << 8)	/* stopped in syscall exit */
#define XTR_INFUNC	(1 << 9)	/* stopped in fuction call */
#define XTR_OTHER	(1 << 10)	/* stopped by some other reason */
#define XTR_CHILDEXITED	(1 << 20)	/* child has called exit() */
#define XTR_STARTCHLD	(1 << 21)	/* xtrace has called start_app */

/* symbol table */
struct symbol {
	vaddr_t va;		/* virtual address in debuggee address space */
	char *name;		/* symbol's name */
	unsigned flags;		/* see below for valid values */
};

#define SYM_REFERENCED	(0x01 << 0)	/* referenced from executable */
#define SYM_SETWATCH	(0x01 << 1)	/* watchpoint can be set */

enum argtype { HEX = 1, INT, STRING, SOCKADDR, MSGHDR, SIGALTSTACK,
	IOCTL, ARGV, ITIMER, FCNTL, TIMEVAL, TIMEZONE, IOVEC, STATFS,
	RTPRIO, STAT, RLIMIT, TIMESPEC, MODSTAT, KLDSTAT, SHEDPARAM,
	SIGSET, SIGACTION, KEVENT };

#define MAXARGS		10
struct syscall {
	int no;					/* syscall's number */
	const char *name;			/* syscall's name */
	int nargs;				/* number of parameters */
	enum argtype args[MAXARGS];		/* MAXARGS parameters max */
};

/* function call data */
struct calldata {
	vaddr_t ip;		/* instruction pointer */
	vaddr_t retval;		/* return value */
	vaddr_t fp;		/* frame pointer */
	vaddr_t rta;		/* return address */
	vaddr_t ofp;		/* previous frame pointer */
	vaddr_t **argv;		/* array of arguments */
};

/* initialize tracer */
extern int xtr_init(struct opt *, char *, char *);

/* find symbol for given virtual address */
extern const struct symbol *xtr_find_symbol(vaddr_t, ptrdiff_t *);

/* get list of dependant ELF binaries */
extern char **xtr_get_elfs(pid_t);

/* build symbol table */
extern int xtr_symtab_build(const struct opt *, char *const *);

/* destroy symbol table */
extern void xtr_symtab_destroy(void);

/* get syscall's character string by number */
extern const struct syscall *xtr_lookup_syscall(int);

/* stop on event of interest */
extern int xtr_stop_app(struct opt *);

/* resume process */
extern int xtr_resume_app(struct opt *);

/* iterate over symbolint xtr_stop_app(const struct opt *);s */
extern struct symbol **xtr_get_next_sym(struct symbol **);

/* read memory content from process's address space */
extern void xtr_read_mem(const struct opt *, vaddr_t, char *, int);

/* get call data by stack pointer */
extern int xtr_get_call_data(struct opt *, struct calldata *);

/* dump current time in format %x.%x (sec, usec) */
extern void xtr_dump_time(void);

/* get the frame pointer */
extern vaddr_t xtr_get_fp(struct opt *);

/* set callback to be called after forking a child */
extern void xtr_before_exec(struct opt *);

/* get the base load address for given ELF object */
extern vaddr_t xtr_get_load_addr(const struct opt *, const char *);

/* dump call */
extern void xtr_dump_call(struct opt *);

#endif /* __XTRACE__H */
